def template_path():
    from dxpy.filesystem import Path
    import dxpy
    p = Path(dxpy.__file__).father


class SnippetMaker:

    @classmethod
    def service(cls, name, path='.'):
        pass

    @classmethod
    def component(cls, name, path='.'):
        from . import snippet
        snippet.Component(name, path).make()



import unittest
from dxpy.task import database


class TestDatabase(unittest.TestCase):
    def setUp(self):
        c = database.config
        c['path'] = ':memory:'
        database.Database.create()

    def tearDown(self):
        database.Database.clear()
        database.config.back_to_default()

    @unittest.skip
    def test_create_engine(self):
        database.Database().get_or_create_engine()

    def test_create_database(self):
        database.Database.create()
        database.Database.create()
        database.Database.clear()

    @unittest.skip
    def test_get_session(self):
        sess = database.Database().session()
        sess.query(database.TaskDB).all()
        sess.close()



import unittest
from dxpy.configs.base import Configs


class TestConfigs(unittest.TestCase):
    def test_basic(self):
        class ConfigsA(Configs):
            _names = ('name1', 'name2')
            default_configs = {'name1': 'value1'}

        a = ConfigsA()
        self.assertEqual(a['name1'], 'value1')

    




from .base import Net


class GlobalStep(Net):
    def __init__(self, config):
        super(__class__, self).__init__(config)
        self.main = tf.Variable(
            0, dtype=tf.int32, trainable=False, name='global_step')

    def old(self):
        self.gs = tf.Variable(
            0, dtype=tf.int32, trainable=False, name='global_step')
        with tf.name_scope('global_step_setter'):
            self.__gs_value = tf.placeholder(dtype=tf.int32, name='gs_value')
            self.__gs_setter = self.gs.assign(self.__gs_value)
        self.add_node('global_step', tensor=self.gs, shape=[])

    def run(self, session, task=None, feeds=None):
        if task is None or task.upper() == "GET":
            return session.run(self.main)
        elif task.upper() == "SET":
            return session.run(self.__gs_setter, feed_dict={self.__gs_setter: feeds['global_step']})
        self.run_unknwn_task(session, task)



from ..graph import Graph
import numpy as np


class Database(Graph):
    """Base class of database
    Database add some special tasks based on graph:
        1. load single
        2. load batch
        3. load next N samples (given N)
        4. load next N samples with batch (given N)
    """

    def __init__(self, name):
        super(__class__, self).__init__(name)

    def load_single(self, feed_dict=None, name=None):
        raise NotImplementedError

    def load_batch(self, feed_dict=None, name=None):
        result = np.zeros(self.c['batch_shape'])
        for _ in range(self.c['batch_size']):
            pass

    def load_N(self, feed_dict=None, name=None):
        raise NotImplementedError

    def load_N_batch(self, feed_dict=None, name=None):
        raise



from dxpy.configs import Configs

from .dataset import Dataset


class ConfigsMNISTDataset(Configs):
    _names = ('path', 'batch_size')
    _default_configs = {'batch_size': 32}


class MNIST(Dataset):
    def __init__(self, config):
        super(__class__, self).__init__(config)



import rx
from rx import Observable
from collections import namedtuple
import re
import os

SINFO = namedtuple('SINFO', ('id', 'part', 'cmd', 'usr',
                             'stat', 'time', 'nodes', 'node_name'))


class TaskInfo:
    def __init__(self, tid, part, cmd, usr, stat, time, nodes, node_name, *args):
        self.id = int(tid)
        self.part = part
        self.cmd = cmd
        self.usr = usr
        self.stat = stat
        self.time = time
        self.nodes = nodes
        self.node_name = node_name

    @classmethod
    def from_squeue(cls, l):
        s = re.sub('\s+', ' ', l).strip()
        items = s.split()
        if not items[0].isdigit():
            return None
        else:
            return cls(*items)


def apply_command(command):
    with os.popen(command) as fin:
        return fin.readlines()


def sbatch(workdir, script_file):
    result = apply_command('cd {dir} && sbatch {file}'.format(
        dir=workdir, file=script_file))
    return sid_from_submit(result[0])


def sid_from_submit(s):
    return int(re.sub('\s+', ' ', s).strip().split(' ')[3])


def squeue():
    return (Observable.from_(apply_command('squeue'))
            .map(lambda l: TaskInfo.from_squeue(l))
            .filter(lambda l: l is not None))


def is_complete(sid):
    if sid is None:
        return False
    return (squeue()
            .subscribe_on(rx.concurrency.ThreadPoolScheduler())
            .filter(lambda tinfo: tinfo.id == sid)
            .count()
            .to_blocking().first()) == 0



from fs.osfs import OSFS
from .model import Directory


def mkdir(d, recreate=True):
    with OSFS('/') as fs:
        fs.makedirs(d.path.rel, recreate=recreate)


def mv(d, path_new):
    with OSFS('/') as fs:
        fs.movedir(d.path_rel, path_new)


def cp(d, path_new, create=True):
    with OSFS('/') as fs:
        fs.copydir(d.path.rel, path_new, create)


def rm(d):
    with OSFS('/') as fs:
        fs.removetree(d.path.rel)



def api_root(version):
    return "/api/v{version}".format(version=version)


def api_path(name, suffix=None, version=None, base=None):
    if base is None:
        base = api_root(version)
    else:
        if base.startswith('/'):
            base = base[1:]
        base = "{root}/{base}".format(root=api_root(version), base=base)
    
    if base.endswith('/'):
        base = base[:-1]    
    if suffix is None:
        return "{base}/{name}".format(base=base, name=name)
    else:
        return "{base}/{name}/{suffix}".format(base=base, name=name, suffix=suffix)


def req_url(name, ip=None, port=None, suffix=None, version=None, base=None):
    return 'http://{ip}:{port}{path}'.format(ip=ip, port=port, path=api_path(name, suffix, version, base))



def summary_var(grads):
    for grad, var in grads:
        if grad is not None:
            tf.summary.histogram(var.op.name + '/gradients', grad)
        for var in tf.trainable_variables():
            tf.summary.histogram(var.op.name, var)


def train_step(tower_grads, optimizer, summary_callback, name='train_step', clipv=None):
    """
    Inputs:
        tower_grads: *list or tuple* of grads
        optimizer: optimizer
    """
    with tf.name_scope(name), tf.device('/cpu:0'):
        average_grads = []
        for grad_and_vars in zip(*tower_grads):
            # Note that each grad_and_vars looks like the following:
            #   ((grad0_gpu0, var0_gpu0), ... , (grad0_gpuN, var0_gpuN))
            grads = []
            for g, _ in grad_and_vars:

                # Add 0 dimension to the gradients to represent the tower.
                expanded_g = tf.expand_dims(g, 0)

                # Append on a 'tower' dimension which we will average over
                # below.
                grads.append(expanded_g)
            # Average over the 'tower' dimension.
            grad = tf.concat(axis=0, values=grads)
            grad = tf.reduce_mean(grad, 0)
            if clipv is not None:
                grad = tf.clip_by_value(grad, -clipv, clipv)
            # Keep in mind that the Variables are redundant because they are shared
            # across towers. So .. we will just return the first tower's pointer to
            # the Variable.
            v = grad_and_vars[0][1]
            grad_and_var = (grad, v)
            average_grads.append(grad_and_var)
        train_op = optimizer.apply_gradients(
            average_grads, global_step=self.gs)

        if summary_callback is not None:
            summary_callback(average_grads)
    return train_op



""" Base class definition """
from collections import defaultdict
import pathlib
import tensorflow as tf
import os
import re
import numpy as np
from collections import ChainMap
from ..utils.general import with_config, ProgressTimer
from ..utils.prints import pp_json, pprint
from ..utils.options import Params
from tqdm import tqdm

from dxpy.collections.dicts import DXDict

# On restrict mode, tensors as allocated to CPU memory if not specified.
RESTRICT_MODE = True


class Net:
    """ Base class of nets.
    Which is used for:
        configs
        # TODO complete doc.
        # TODO deprecate device type, use auto inference by nb_gpus
    Level0:
        node: {node name: tensor}
        nodes_name_proxy: {node name: data name}
        hypers: {node name: values/ndarray}

        train_tasks: list of string, sub task names
        train_vars: {train sub task name: list of vars}
        losses: {train sub task name: tf.tensor}
        optimizers: {train sub task name: tf.train.xxxOptimizer}
        metrics: {train sub task name: tf.tensor}
        grads: {train sub task name: list/tuple of tf.tensors}
        train_steps: {train sub task name: tf.op}

    Level1:
        run_ops: { task names : *dict* of tf.ops to run }
        feed_dict: { task names : list of name of nodes to be feeded }

        run these tasks by calling run.
    Level2:
        datasets: {task name: dataset}
        data - net couple:
            data {name: np.ndarray}
            node {name: tf.tensor}
            nodes_name_proxy {name_in_node: name_in_data}
        task - run_op, feed_dict coutple:
            name: task name
            run_op {name: tf.tensors/tf.ops}
            feed_dict {name: node_names}
        lr - optimizer - train_step couple:
            loss {name: tf.tensor}
            optimizer {name: optimizer}
            train_step {name: tf.op}
    """

    def __init__(self, config):
        """
        Inputs:
            devices_type: 'auto' or 'gpus'. If 'gpus', losses and grads are lists, [0] cpu [1-n] gpus.
        """
        from . import train
        self.config = config

        # model external nodes
        self.nodes = DXDict()
        self.subnets = DXDict()

        # Variables and constants (stored in hypers)
        self.hypers = dict()

        #global step
        self.global_step = train.GlobalStep(config['train/global_step'])

        # keep prob
        self.kp, _ = self.add_node('keep_prob', shape=[])

        # training switch
        self.training, _ = self.add_node('training', shape=[], dtype=tf.bool)

        # learning rates
        if not isinstance(lr, dict):
            lr = {'main': lr}
        self.params['lr'] = lr
        self.lr = dict()

        for k in self.params['lr']:
            name = 'lr/' + k
            self.lr[k], _ = self.add_node(name, shape=[])
            self.hypers[name] = self.params['lr'][k]

        # train sub tasks
        self.train_tasks = ['main']
        self.train_vars = dict()
        self.losses = dict()
        self.grads = dict()
        self.metrices = dict()
        self.optimizers = dict()
        self.train_steps = dict()

        # Key by task
        self.summary_ops = dict()

        self.feed_dict = dict()
        self.feed_dict['default'] = list()
        for k in self.params['lr']:
            self.feed_dict['default'].append('lr/' + k)
        self.run_op = {
            'default': {'global_step': self.gs}
        }

        # Debug tensors
        self.debug_tensor = dict()

        # Key with 'train' and 'test'
        self.dataset = dict()

        # Session and managers
        self.sess = None
        self.saver = None
        self.summary_writer = None
        self.supervisor = None

        # Quick flags
        self.is_multi_gpu = self.p.nb_gpus > 1

    # Helper functions for constructing net:

    def add_node(self, name=None, tensor=None, is_gpu=False, dtype=None, shape=None, init_value=None, export_name=None):
        """ Add node for net. Nodes can be regard as API of Net.

        Rule:
            All tensors which is feedable need to be added
            Which implies all placeholders should be added via add_node.

            All outputs should be added to nodes.

        Inputs:
            name: node name, if is None, use tensor.name
            shape: node shape/tensor shape, if is `None`, use tensor.shape
            tensor: tensor to add to nodes. If is `None`, a placeholder will be created.
            is_multi_gpu: split tensor into multiple partial mini-batches.
        Returns:
            tensor: main part of tensor
            tensors_gpu: list of splited tensors
        Raises:
            ValueError
        """
        from . import tensor
        from .tensor import GPUTensorCreator
        tensor, tensors_gpu = GPUTensorCreator.create(
            tensor, name, shape, dtype, nb_gpu)
        self.nodes[name] = tensor
        if proxy_name is not None:
            self.nodes_name_proxy[name] = proxy_name
        return tensor, tensors_gpu

    @property
    def as_tensor(self):
        return self.main

    def __getitem__(self, name):
        from dxpy.filesystem.path import Path
        name = Path
        if name in self.nodes:
            return self.nodes[name]
        else:
            names = name.split('/')

            return self.nodes[names[0]]['/'.join(names[1:])]
        return None

    def run(self, feed_dict):
        pass

    def run_unknwn_task(self):
        pass

    def resolve_feed_dict(self):
        pass

    def init(self):
        pp_json(self.params, self.params['name'] + " PARAMS:")
        self._set_model()
        self._set_train()
        self._set_saver()
        self._set_sesssv()
        self._set_summary()
        self._set_task()
        pp_json(self.run_op, "RUN OPS:")
        pp_json(self.feed_dict, "FEED DICTS:")

    def _set_model(self):
        """ Construct model(s) for net.
        To construct:
            node: {node name: tensor}
            nodes_name_proxy: {node name: data name}

            losses: {sub task name: tf.tensor or list of tf.tensors}                    
        """
        pass

    def _set_task(self):
        """ Constrcut tasks for net.
        Tasks like train, evaluate, summary, predict, etc.
        To construct:
            run_op: { task names : *dict* of tf.ops to run }
            feed_dict: { task names : list of name of nodes to be feeded }
        """
        pass

    def _get_optimizer(self, sub_task_name):
        optim = self.optimizers.get(sub_task_name)
        if optim is None:
            name = self.p.optimizer_name.get(sub_task_name)
            if name in ['Adam', 'adam']:
                optim = tf.train.AdamOptimizer(self.lr[sub_task_name])
            elif name in ['RMSProp', 'rmsprop']:
                optim = tf.train.RMSPropOptimizer(self.lr[sub_task_name])
        return optim

    def _get_train_vars(self, sub_task_name):
        vars_list = self.train_vars.get(sub_task_name)
        return vars_list

    def _set_saver(self):
        # TODO: partial saver and loader.
        self.saver = tf.train.Saver()
        pass

    # def _set_sesssv(self):
    #     # sv_para = {'summary_op': None}
    #     # sms = self.params.get('save_model_secs')
    #     # if sms is not None:
    #     #     sv_para['save_model_secs'] = sms
    #     # if self.params['load_step'] is not None:
    #     #     sv_para['init_fn'] = self.load
    #     # sv_para['saver'] = self.saver
    #     # self.supervisor = tf.train.Supervisor(**sv_para)
    #     if self.params['is_show_device_placement']:
    #         config = tf.ConfigProto(log_device_placement=True)
    #     else:
    #         config = tf.ConfigProto()
    #     config.gpu_options.allow_growth = True
    #     self.sess = tf.Session(config=config)
    #     # self.saver = self.supervisor.saver

    def _set_summary(self):
        self.summary_writer = dict()
        for k in self.p.summary_modes:
            path_log = str(pathlib.Path(self.params['log_dir']) / k)
            self.summary_writer[k] = tf.summary.FileWriter(
                path_log, self.sess.graph)

        # Add summary service
        if self.params['summary_type'] == 'time':
            self.supervisor.loop(
                self.params['summary_freq'], self.summary_auto)

    # level-1 api
    def run(self, task, datas=None, verbose=0):
        """
        """
        run_ops = self.run_op[task]
        true_feed_dict = list(self.feed_dict['default'])
        true_feed_dict += self.feed_dict[task]
        if verbose >= 1:
            pp_json({'run_ops': run_ops, 'to_feed': true_feed_dict},
                    "RUN TASK: " + task)
        feed_dict = dict()
        for k in true_feed_dict:
            k_true = self.nodes_name_proxy.get(k, k)
            datas_all = ChainMap(datas, self.hypers)
            feed_dict[self.nodes[k]] = datas_all.get(k_true)
            if feed_dict[self.nodes[k]] is None:
                msg_format = r"Node: {0} with proxy name {1} was not found in hyers {2} and datas {3}"
                msg_content = (k, k_true, list(
                    self.hypers.keys()), list(datas.keys()))
                raise ValueError(msg_format.format(*msg_content))
        out = self.sess.run(run_ops, feed_dict=feed_dict)
        return out

    def partial_fit(self, data, sub_task=None, verbose=0):
        """ features are dict of mini batch numpy.ndarrays """
        task = 'train'
        if sub_task is not None:
            task = '{0}/{1}'.format(task, sub_task)
        true_data = dict()
        true_data['training'] = True
        true_data['keep_prob'] = self.p.keep_prob
        true_data.update(data)
        return self.run(task, true_data, verbose)

    def predict(self, data, sub_task=None, verbose=0):
        """ predict a mini-batch """
        task = 'predict'
        if sub_task is not None:
            task = '{0}/{1}'.format(task, sub_task)
        true_data = dict()
        true_data['training'] = False
        true_data['keep_prob'] = 1.0
        true_data.update(data)
        return self.run(task, true_data, verbose)

    def evaluate(self, data, sub_task=None, verbose=0):
        task = 'evaluate'
        if sub_task is not None:
            task = '{0}/{1}'.format(task, sub_task)
        true_data = dict()
        true_data['training'] = False
        true_data['keep_prob'] = 1.0
        true_data.update(data)
        return self.run(task, true_data, verbose)

    def dump(self, task, datas, save_name=None, verbose=0):
        result = self.run(task, datas, verbose)
        if save_name is None:
            save_name = 'dump%d' % (self.get_global_step())
        np.savez(save_name, **result)

    def summary(self, data, mode=None, sub_task=None, verbose=0):
        if mode is None:
            mode = 'train'
        task = 'summary'
        if sub_task is not None:
            task = '{0}/{1}'.format(task, sub_task)
        true_data = dict()
        true_data['training'] = False
        true_data['keep_prob'] = 1.0
        true_data.update(data)
        results = self.run(task, true_data, verbose)
        step = self.get_global_step()
        for k in self.run_op['summary'].keys():
            self.summary_writer[mode].add_summary(results[k], global_step=step)
        self.summary_writer[mode].flush()
        return results

    def set_dataset(self, name, dataset):
        self.dataset[name] = dataset

    # level-2 api
    def train_kernel(self, sub_task):
        ss = next(self.dataset[sub_task])
        res = self.partial_fit(ss, sub_task)
        msg = "LOSS=%6e, STEP=%5d" % (res['loss'], res['global_step'])
        return res, msg

    def set_dataset_auto(self, dataset_train, dataset_test):
        self.dataset['main'] = dataset_train
        self.dataset['train'] = dataset_train
        self.dataset['test'] = dataset_test

    def _train_sub_task_schadule(self, sub_task=None):
        if sub_task is None:
            sub_task = 'main'
        while True:
            yield sub_task

    def train(self, sub_task=None, steps=None, phase=1, decay=2.0, warmup=False):
        """ high level train.
        Inputs:
            sub_task: (Optional), if not None, run task train/sub_task
            steps: int or list of ints, steps per phase.
            phase: int. If steps is int, construct steps = [steps] * phase
            decay: learning rate decay per phase
            TODO: check warmup.
        """
        if not isinstance(steps, (list, tuple)):
            steps = [steps] * phase
        total_step = sum(steps)

        task_gen = self._train_sub_task_schadule(sub_task)

        if warmup:
            # warmup
            pt = ProgressTimer(total_step)
            cstep = 0
            lr_bak = self.p.lr['main']
            warming_up_phase = 10
            self.reset_lr(decay=2.0**warming_up_phase)
            pp_json(self.params, self.params['name'] + " PARAMS:")
            warmup_step = self.params['warmup_step']
            for i in range(warming_up_phase):
                for j in range(warmup_step):
                    _, msg = self.train_kernel(next(task_gen))
                    cstep += 1
                    pt.event(cstep, "[WARM]" + msg)
                self.reset_lr(decay=0.5)

        pt = ProgressTimer(total_step)
        cstep = 0
        for idx, sp in enumerate(steps):
            for i in range(sp):
                if warmup and warming_up > 0:
                    if warmup_step <= 0:

                        warming_up -= 1
                        warmup_step = warmup_step = self.params['warmup_step']
                    else:
                        warmup_step -= 1

                # if res['loss'] > 1e-2:
                #     self.dump(ss)
                cstep += 1
                _, msg = self.train_kernel(next(task_gen))
                pt.event(cstep, msg)
                if self.params['summary_type'] == 'step':
                    if i % self.params['summary_freq'] == 0 and i > 0:
                        self.summary_auto()
                if i % self.params['save_freq'] == 0 and i > 0:
                    self.save()

            # force save at end of each phase
            self.save()

            # decay lr at end of phase except the last one.
            if not idx == len(steps) - 1:
                self.reset_lr(decay=decay)

    def predict_auto(self, data, sub_task=None, batch_size=None, **kwargs):
        """ predict a large tensor, automatically seperate it into mini-batches. """
        nb_sample = None
        if batch_size is None:
            batch_size = self.p.batch_size
        for k in data.keys():
            if nb_sample is None:
                nb_sample = data[k].shape[0]
            else:
                if nb_sample != data[k].shape[0]:
                    raise ValueError("Wrong data shape.")
        nb_blocks = nb_sample // batch_size + 1
        preds = []
        pt = ProgressTimer(nb_blocks)
        for i in range(nb_blocks):
            skip = False
            data_block = dict()
            for k in data.keys():
                i_start = i * batch_size
                i_end = min([(i + 1) * batch_size, nb_sample])
                if i_start >= i_end:
                    skip = True
                    break
                data_block[k] = data[k][i_start:i_end, ...]
            if not skip:
                preds.append(self.predict(data_block))
            pt.event(i)
        results = dict()
        for k in preds[0].keys():
            results[k] = []
        for item in preds:
            results[k].append(item[k])
        for k in results.keys():
            results[k] = np.concatenate(results[k], 0)
        return results

    def summary_auto(self):
        result = dict()
        for k in self.summary_writer.keys():
            ss = self.dataset[k].sample()
            result[k] = self.summary(ss, mode=k)
        return result




from ..model.base import Net


def default_path():
    import os
    from dxpy.filesystem import Path, Directory
    path_env = os.environ.get('PATH_DATASETS')
    if path_env is not None:
        dataset_dir = Directory(Path(path_env))
        if dataset_dir.exists():
            return dataset_dir.path
    return None


class Dataset(Net):
    """
    Subnet for loading datasets.    
    """

    def __init__(self, name):
        super(__class__, name).__init__(self)

    def _initialize(self):
        pass

    def _finalize(self):
        pass

    def __enter__(self):
        self._initialize()

    def __exit__(self, type, value, traceback):
        self._finalize()

    def run(self, session=None, task=None, feeds=None):
        return self.batch()

    def single(self, session=None, task=None, feeds=None):
        pass

    def batch():
        pass


class DatasetHDF5(Net):
    def __init__(self, name):
        super(__class__, name).__init__(self)

    def run(self, session=None, task=None, feeds=None):
        pass



import copy
from dxpy.collections.dicts import DXDict


class Configs(DXDict):
    """
    A configs object is a *imutable* dict like collection. elements of the collection can be one of the following:
    #. basic types (str, int, bool)
    #. list or dict of basic types
    #. Configs object
    #. list or dict of Configs
    """
    yaml_tag = '!configs'
    _names = tuple()
    _default_configs = {}

    def __init__(self, *args, default_dict=None, **kwargs):
        if default_dict is None:
            default_dict = DXDict(self._default_configs)
        else:
            default_dict = default_dict.apply_default(default_dict)
        super(__class__, self).__init__(
            *args, **kwargs, default_dict=default_dict)

    @classmethod
    def names(cls):
        """ All keys of a configs class.
        """
        return cls._names


from dxpy import serialization
serialization.register(Configs)



from collections import namedtuple
import copy
import rx
import dask
import sys
import os
import dxpy.slurm as slurm
from ..model import task as taskpy
from ..model import creators
from .. import interface

NB_THREADS = 5
THREAD_POOL = rx.concurrency.ThreadPoolScheduler(NB_THREADS)


class Workers:
    @classmethod
    def is_complete(cls, task, *args):
        return task.state == taskpy.State.Complete

    @classmethod
    def on_this_worker(cls, task):
        return task.worker == cls.WorkerType

    @classmethod
    def plan(cls, task):
        raise NotImplementedError

    @classmethod
    def run(cls, task, stdout=None, stderr=None):
        print('RUN CALLED with task')
        print(task)
        if stdout is None:
            stdout = sys.stdout
        if stderr is None:
            stderr = sys.stderr
        (rx.Observable.from_([task])
         .subscribe_on(THREAD_POOL)
         .map(cls.plan)
         .subscribe(on_next=lambda r: print(r, file=stdout),
                    on_error=lambda e: print(e, file=stderr)))


class NoAction(Workers):
    WorkerType = taskpy.Worker.NoAction

    @classmethod
    def plan(cls, t, *args):
        print("NO ACTION PLAN CALLED")
        t = interface.mark_start(t)
        t = interface.mark_complete(t)
        return 'NoAction of task id: {} done.'.format(t.id)


def sbatch_command(workdir, file):
    cmd = 'cd {0}'.format(workdir)
    if file is not None:
        cmd += ' && sbatch {0}'.format(file)
    return cmd


def normal_command(workdir, command):
    return 'cd {0} && {1}'.format(workdir, command)


class Slurm(Workers):
    WorkerType = taskpy.Worker.Slurm

    @classmethod
    def is_complete(cls, task, *args):
        return slurm.is_complete(task.data.get('sid'))

    @classmethod
    def plan(cls, task, *args):
        # why?
        # TODO: FIX
        # print(args)
        # raise ValueError('task {0}, args: {1}'.format(task, args))
        # def srun_command(workdir, command):
        #     return 'cd {0} && srun {1}'.format(workdir, command)

        task = interface.mark_start(task)

        # if isinstance(task, templates.TaskCommand):
        #     command = task.command(srun_command)
        if not task.type == taskpy.Type.Script:
            raise TypeError(
                'Slurm worker only support TaskScript tasks, got: {!r}.'.format(task))
        command = sbatch_command(task.workdir, task.data.get('file'))
        with os.popen(command) as fin:
            result = fin.readlines()[0]
        sid = slurm.sid_from_submit(result)
        task.data.update({'sid': sid})
        interface.update(task)
        return result


class MultiThreding(Workers):
    WorkerType = taskpy.Worker.MultiThreading

    @classmethod
    def plan(cls, task):
        # TRT = namedtuple('TaskResultTuple', ('task', 'res'))
        task = interface.mark_start(task)
        if task.type == taskpy.Type.Command:
            with os.popen(normal_command(task.workdir, task.data['command'])) as fin:
                result = fin.readlines()
        task = interface.mark_complete(task)
        return result


WORKERS = [NoAction, MultiThreding, Slurm]


def get_workers(task):
    for w in WORKERS:
        if w.on_this_worker(task):
            return w

# class Dask(Workers):
#     WorkerType = taskpy.Worker.Dask
#     NB_PROCESSES = 5

#     @classmethod
#     def run_plan(cls, task):
#         (rx.Observable.just(task.workers.num_workers)
#          .map(lambda i: dask.delayed(task.run)(i_worker=i))
#          .to_list()
#          .map(lambda l: dask.bag.from_delayed(l))
#          .map(lambda b: b.compute(num_workers=NB_PROCESSES)))



import pathlib
import urllib
import yaml


class PathError(Exception):
    pass


class NonePathError(PathError, TypeError):
    pass


class NotUrlPathError(PathError, ValueError):
    pass


class IsUrlPathError(PathError, ValueError):
    pass


class NotValidPathError(PathError, ValueError):
    pass


def _is_url_path(path, is_strict=False):
    if not isinstance(path, str):
        raise TypeError
    if '/' in path and not "%252F" in path:
        return False
    if not '/' in path and "%252F" in path:
        return True
    if not '/' in path and not "%252F" in path:
        return not is_strict
    raise NotValidPathError(path)


def _url_quote(path):
    if _is_url_path(path, is_strict=True):
        raise IsUrlPathError(path)
    qpath = urllib.parse.quote_plus(str(pathlib.Path(path).absolute()))
    qqpath = urllib.parse.quote_plus(qpath)
    return qqpath


def _url_unquote(path_url):
    if not _is_url_path(path_url):
        raise NotUrlPathError
    qpath = urllib.parse.unquote_plus(path_url)
    path = urllib.parse.unquote_plus(qpath)
    return path


class Path(yaml.YAMLObject):
    """ Unified path for posix/windows/url paths.
    This class is an 'abstract' path object, which means it only normalize its representation for different platform,
    but there is not any relation to the true file system. Thus this class does **NOT** provide methods like `exists()` or
    `is_dir()`, please refer `dxpy.file_syste.file.File` for these functions.
    """
    yaml_tag = '!path'

    def __init__(self, path, is_url_path=None, ):
        """
        -`is_url_path`: a bool to specify if given path is a url path. Default is `None` which requires automatic determination.
        """
        if isinstance(path, Path):
            self.path = path.path
            return
        if isinstance(path, pathlib.Path):
            self.path = path
            return
        if path is None:
            raise NotValidPathError
        if is_url_path is None:
            is_url_path = _is_url_path(path)
        if is_url_path:
            self.path = pathlib.Path(_url_unquote(path)).absolute()
        else:
            self.path = pathlib.Path(path).absolute()

    def __str__(self):
        return '<dxpy.file_system.path.Path object with path: {abs}>'.format(abs=self.abs)

    def __truediv__(self, name):
        return Path(self.path / name)

    def parts(self):
        return self.path.parts

    @property
    def parent(self):
        return Path(self.path.parent, is_url_path=False).abs

    @property
    def name(self):
        return self.path.name

    @property
    def url(self):
        return _url_quote(self.abs)

    @property
    def abs(self):
        return str(self.path.absolute())

    @property
    def suffix(self):
        return self.path.suffix

    def check_exists(self, file_system):
        return file_system.exists(self.path)

    @property
    def brief(self):
        return {
            'name': self.name,
            'path': self.abs
        }

    @property
    def detail(self):
        result = self.brief
        result.update({
            'parent': self.parent,
            'url': self.url,
            'suffix': self.suffix,
            'parts': self.parts()
        })
        return result

    @classmethod
    def to_yaml(cls, dumper, data):
        return yaml.ScalarNode(cls.yaml_tag, data.abs)

    @classmethod
    def from_yaml(cls, loader, node):
        return Path(loader.construct_scalar(node))

    def __eq__(self, path):
        return self.abs == path.abs

# def _path_representer(dumper, data):
#     return dumper.represent_scalar('!path', data.abs)


# def _path_constructor(loader, node):
#     value = loader.construct_scalar(node)
#     return Path(value)


# # yaml.add_representer(Path, _path_representer)
# yaml.add_constructor('!path', _path_constructor)



"""
"""
# Level-0:


# class FilesBatchWorker:
#     from .. import service, model

#     def __init__(self, fs=None, include_pattern=None, exclude_pattern=None, depth=1):
#         from fs.osfs import OSFS
#         fs = fs or OSFS('.')
#         self.fs = fs
#         self.filter = model.FilesFilter(
#             include_pattern, exclude_pattern, depth)

#     def ls(self):
#         """
#         List all files matches filters
#         """
#         self.service.Mapper.ls(self.fs, self.filter)

#     @classmethod
#     def merge(fs, filters, target=None, method=None):
#         """
#         Merge all files matches filters into one.
#         """
#         pass

#     @classmethod
#     def map(fs, filters, callback, args, kwargs):
#         """
#         Map callback to files like:
#             for f in valid_files:
#                 callback(*args, filename=f, **kwargs)
#         """
#         pass

#     @classmethod
#     def clear(fs, filters):
#         pass


# class DirsBatchWorker:
#     @classmethod
#     def ls(fs, filters, show_size=False):
#         """
#         List all files matches filters, show their properties.
#         """
#         pass

#     @classmethod
#     def merge(cls, fs, filters, target=None, is_check=True, is_auto_rename=True, rename_pattern=r'{name}.{id}.{suffix}'):
#         """
#         Merge directories, provide check, auto rename options.
#         """
#         pass

#     @classmethod
#     def clear(cls, fs, include_filters=None, exclude_filters=None, depth=1, dryrun=False):
#         from ..service import Mapper
#         from ..model import DirsFilter
#         from functools import partial
#         dirs = DirsFilter(include_filters=include_filters,
#                           exclude_filters=exclude_filters, depth=depth)
#         if dryrun:
#             print("The Following directories are going to be removed")
#             Mapper.map(fs, dirs, lambda d: print('[REMOVE]', d))
#         else:
#             Mapper.map(fs, dirs, fs.removetree)


# # Level-1
# def list_dirs(target, pattern):
#     pass


# def clear_dirs(target, include_filters, exclude_filters=None, depth=1, dryrun=False):
#     from dxpy.filesystem.smartfs import FS
#     with FS(target) as fs:
#         DirsBatchWorker.clear(fs, include_filters,
#                               exclude_filters, depth, dryrun)


from ..model.filters import DirectoriesFilter, FilesFilter
from ..service import Mapper, files_in_directories
from ..exceptions import *



"""
Graph with dipendencies
"""

import networkx as nx
from dxpy.exceptions.checks import assert_same_length


class DenpensGraph:
    def __init__(self, nodes=None, depens=None):
        """
            n1 depens on n2 if n2 in n1.out_nodes.
        """
        assert_same_length((nodes, depens), ('nodes', 'depens'))
        self.g = nx.DiGraph()
        self.g.add_nodes_from(nodes)
        for i, ds in enumerate(depens):
            self.add_depens(self.g.nodes()[i], ds)

    def add_node(self, n):
        self.g.add_node(n)

    def remove_node(self, n):
        self.g.remove_node(n)

    def add_depens(self, node, depens):
        if depens is None or depens == [None]:
            return
        if not isinstance(depens, (list, tuple)):
            self.g.add_edge(node, depens)
            return
        for d in depens:
            self.g.add_edge(node, d)

    def draw(self):
        nx.draw(self.g, with_labels=True)

    def is_free(self, node):
        return self.g.out_degree(node) == 0

    def is_root(self, node):
        return self.g.in_degree(node) == 0

    def free_nodes(self):
        return [n for n in self.g if self.is_free(n)]

    def root_nodes(self):
        return [n for n in self.g if self.is_root(n)]

    def is_depens_on(self, node_succ, node_depended):
        return node_depended in self.g.successors(node_succ)

    def dependencies(self, n):
        return self.g.successors(n)

    def nodes(self):
        return self.g.nodes()

    def __len__(self):
        return len(self.g)



import click


@click.group()
def files():
    pass


@files.command()
def sum():
    pass


@files.command()
def cat():
    pass


@files.command()
def hadd():
    pass


@files.group()
def clears():
    """
    Quick cleaner of prefab utiles
    """


@clears.option("--name", "-n", type=str, help="Name of prefab utiles.")
def pyc():
    pass


@click.group()
def dirs():
    pass


@dirs.command()
def merge():
    pass


@dirs.command()
def sbatch():
    pass



from flask import Flask, request, jsonify, url_for, make_response
from flask_restful import Resource, Api
import requests

from dxpy.tasks.task import TaskService, TaskSleep


class TaskResource(Resource):
    def get(self, id):
        return str(TaskService.get(id=int(id)))

    def put(self, id):
        task = TaskService.put(**request.form)
        return str(task)


class TasksResource(Resource):
    def get(self):
        return [str(t) for t in TaskService.get_all()]

    def post(self):
        kwargs = {k: request.form[k] for k in request.form}
        task = TaskService.new(**kwargs)
        return str(task), 201

# Server logic


class TaskServer:
    """
    """

    def __init__(self, config_file=None):
        self.app = Flask(__name__)
        self.api = Api(self.app)
        self.api.add_resource(TaskResource, '/api/v0.1/task/<int:id>')
        self.api.add_resource(TasksResource, '/api/v0.1/tasks')

    def start(self, host=None, port=5000, debug=True):
        self.app.run(host=host, port=port, debug=debug)


server = TaskServer()

if __name__ == "__main__":
    server.start()



import rx
from fs import copy, path


def files_in_directories(fs, include_dirs, include_files, depth=0, exclude_dirs=None, exclude_files=None):
    from .model.filters import FilesFilter, DirectoriesFilter
    results = list()
    (DirectoriesFilter(include_dirs, exclude_dirs, depth).obv(fs)
     .flat_map(lambda p: FilesFilter(include_files, exclude_files).obv(fs, p))
     .subscribe(results.append))
    return results


class Mapper:
    @classmethod
    def ls(cls, fs, paths, infos=None):
        # TODO convert to true observalbes
        infos = infos or ['syspath']
        result_dct = dict()
        for k in infos:
            if k == 'syspath':
                result_dct.update({k: [fs.syspath(p) for p in paths]})
            elif k == 'size':
                result_dct.update({k: [fs.getdetails(p).size for p in paths]})
        results = zip(*(result_dct[k] for k in result_dct))
        # return rx.Observable.from_(results)
        return results

    @classmethod
    def map(cls, fs, paths, callback):
        return [callback(fs, p) for p in paths]

    @classmethod
    def broadcast(cls, fs, paths, path_content):
        """
        Copy content to all paths.
        """
        return [copy.copy_file(fs, path_content, fs, p) for p in paths]

    @classmethod
    def broadcast_directory(cls, fs, paths, path_content):
        name = path.basename(path_content)
        return [copy.copy_file(fs, path_content, fs, path.join(p, name)) for p in paths]


class Reducer:
    @classmethod
    def cat(cls, fs, paths):
        """
        Inputs:
            files: a list/observable of File.
        """
        pass



from fs.osfs import OSFS
from ..path.model import Path


class File:
    def __init__(self, path, load_depth=0):
        self.path = Path(path)
        if load_depth > 0:
            self.load(load_depth)
        else:
            self.contents = None

    @property
    def exists(self):
        with OSFS('/') as fs:
            return fs.exists(self.path.rel)

    def load(self, depth):
        self.contents = None
        if depth < 0:
            return
        if not self.exists:
            raise FileNotFoundError(self.path.abs)
        if depth == 0:
            return
        with OSFS('/') as fs:
            with fs.open(self.path.abs, 'rb') as fin:
                self.contents = fin.read()

    def save(self, data):
        with OSFS('/') as fs:
            with fs.open(self.path.abs, 'wb') as fout:
                self.contents = fout.write(data)

    def to_serializable(self):
        try:
            cont = self.contents.decode() if self.contents else None
        except UnicodeDecodeError:
            cont = '!!binary' + str(self.contents)
        return {'path': self.path.abs,
                'name': self.path.name,
                'is_dir': False,
                'contents': cont}

    def __str__(self):
        import json
        return json.dumps(self.to_serializable(), sort_keys=True, separators=(',', ':'), indent=4)



# Auto generated web.py
from flask import Flask, make_response, request, Response
from flask_restful import Resource, reqparse, Api
from . import api

# TODO: Implementation
raise NotImplementedError


class BatchResource(Resource):
    def get(self, id):
        try:
            return Response(api.read(id), 200, mimetype="application/json")
        except Exception as e:
            return str(e), 404

    def put(self, id):
        try:
            obj = request.form['obj']
            return Response(api.update(task), 201, mimetype="application/json")
        except Exception as e:
            return str(e), 404

    def delete(self, id):
        try:
            return Response(api.delete(id), 200, mimetype="application/json")
        except Exception as e:
            return str(e), 404


class BatchsResource(Resource):
    import json

    def get(self):
        try:
            objs = []
            api.get_all().subscribe(lambda o: objs.append(o))
            return Response(__class__.json.dumps(objs), 200, mimetype="application/json")
        except Exception as e:
            return str(e), 404

    def post(self):
        obj = request.form['obj']
        result = api.create(obj)
        return Response(__class__.json.dumps({'result': result}), 201, mimetype="application/json")


def add_api(api, root):
    from .config import c
    api.add_resource(BatchResource, c.get_url_endpoint(root))
    api.add_resource(BatchsResource, c.get_url_endpoint(root, plural=True))


def launch_server():
    from .config import c
    app = Flask(__name__)
    api = Api(app)
    add_api(api, c.get_root())
    app.run(host=c.ip, port=c.port, debug=c.debug)



    # def _set_sesssv(self):
    #     sv_para = {'summary_op': None}
    #     sms = self.params.get('save_model_secs')
    #     if sms is not None:
    #         sv_para['save_model_secs'] = sms
    #     if self.params['load_step'] is not None:
    #         sv_para['init_fn'] = self.load
    #     sv_para['saver'] = self.saver
    #     self.supervisor = tf.train.Supervisor(**sv_para)
    #     if self.params['is_show_device_placement']:
    #         config = tf.ConfigProto(log_device_placement=True)
    #     else:
    #         config = tf.ConfigProto()
    #     config.gpu_options.allow_growth = True
    #     self.sess = self.supervisor.prepare_or_wait_for_session(config=config)
    #     # self.saver = self.supervisor.saver


from fs.osfs import OSFS
from ..exceptions import DirectoryNotFoundError
from ..path.model import Path
from ..file.model import File

import yaml


class Directory:
    yaml_tag = '!directory'

    def __init__(self, path, load_depth=0, file_system_cls=None):
        self.path = Path(path)
        if load_depth > 0:
            self.load(load_depth)
        else:
            self.children = None
        self.fs_cls = file_system_cls or OSFS

    def load(self, depth):
        self.children = None
        if depth < 0:
            return
        if not self.exists:
            raise DirectoryNotFoundError(self.path.abs)
        if depth == 0:
            return
        self.children = []
        with OSFS('/') as fs:
            for f in fs.listdir(self.path.rel):
                fp = self.path / f
                if fs.isdir(fp.rel):
                    self.children.append(Directory(fp, depth - 1))
                else:
                    self.children.append(File(fp, 0))

    @property
    def exists(self):
        with OSFS('/') as fs:
            return fs.exists(self.path.rel)

    def ensure(self):
        if not self.exists:
            with OSFS('/') as fs:
                fs.makedirs(self.path.rel)

    def to_serializable(self):
        return {'path': self.path.abs,
                'name': self.path.name,
                'is_dir': True,
                'children': [c.to_serializable() for c in self.children] if self.children is not None else None
                }

    def __str__(self):
        import json
        return json.dumps(self.to_serializable(), sort_keys=True, separators=(',', ':'), indent=4)


# def directory_representer(dumper, data):
#     return dumper.represent_scalar(Path.yaml_tag, data.abs)


# def directory_constructor(loader, node):
#     value = loader.construct_scalar(node)
#     return Path(value)


# yaml.add_representer(Directory, directory_representer)
# yaml.add_constructor(Directory.yaml_tag, directory_constructor)



import rx


class BaseFilter:
    def __init__(self, include_filters=None, exclude_filters=None):
        """
        Priority: exclude > include
        depth: int, depth of scanning.
            1 : only subdirs
            0 : only current dir
            -1: unlimited
        """
        super(__class__, self).__init__()
        self.include_filters = include_filters or []
        self.exclude_filters = exclude_filters or []

    def _apply(self, fs, path):
        raise NotImplementedError

    @classmethod
    def _filter_paths(cls, fs, path, files=None, dirs=None, exclude_files=None, exclude_dirs=None):
        infos = list(fs.filterdir(path,
                                  files=files,
                                  exclude_files=exclude_files,
                                  dirs=dirs,
                                  exclude_dirs=exclude_dirs))
        return (rx.Observable.from_(infos)
                  .map(lambda info: info.make_path(path)))

    @classmethod
    def _filter_files(cls, fs, path, include_filters, exclude_filters):
        return cls._filter_paths(fs, path, files=include_filters, exclude_files=exclude_filters, exclude_dirs=['*'])

    @classmethod
    def _filter_dirs(cls, fs, path, include_filters, exclude_filters=None):
        return cls._filter_paths(fs, path, exclude_files=['*'], dirs=include_filters, exclude_dirs=exclude_filters)

    def obv(self, fs, path='.'):
        return self._apply(fs, path)

    def lst(self, fs, path='.'):
        result = []
        self.obv(fs, path).subscribe(result.append)
        return result


class FilesFilter(BaseFilter):
    def __init__(self, include_filters=None, exclude_filters=None):
        super(__class__, self).__init__(include_filters, exclude_filters)

    def _apply(self, fs, path, depth=None):
        return self._filter_files(
            fs, path, self.include_filters, self.exclude_filters)


class DirectoriesFilter(BaseFilter):
    def __init__(self, include_filters=None, exclude_filters=None, depth=0):
        super(__class__, self).__init__(include_filters, exclude_filters)
        self.depth = depth

    @classmethod
    def _next_depth(cls, depth):
        return depth - 1 if depth > 0 else -1

    @classmethod
    def _need_recursive(cls, depth):
        return depth > 0 or depth == -1

    def _apply(self, fs, path, depth=None):
        if depth is None:
            depth = self.depth
        result = self._filter_dirs(
            fs, path, self.include_filters, self.exclude_filters)
        if self._need_recursive(depth):
            sub_results = (result.flat_map(
                lambda p: self._apply(fs, p, self._next_depth(depth))))
            result = rx.Observable.merge(result, sub_results)
        return result



import sys
class ExceptionHook:
    instance = None

    def __call__(self, *args, **kwargs):
        if self.instance is None:
            from IPython.core import ultratb
            self.instance = ultratb.FormattedTB(mode='Plain',
                                                color_scheme='Linux', call_pdb=1)
        return self.instance(*args, **kwargs)


def enter_debug():
    sys.excepthook = ExceptionHook()



def parse_configs(func, *args, configs_filename=None, **kwargs):
    """ Analysic func args and inputs/JSON file.

    Inputs:

        * args: tuple of positional arguments,                
        * configs_filename: filename or list of filenames of JSON file(s), *which may
          or may not present in original function's argument list*.
        * kwargs: dict of keyword arugments.

    Returns:
        1. a tuple, positional arguments,        
        2. a dict, keyword arugments.

    Raises:

    """
    # TODO: refine doc, add test.
    if isinstance(configs_filename, str):
        configs_filename = [configs_filename]
    if configs_filename is None:
        configs_filename = []
    json_dicts = []
    for fn in configs_filename:
        with open(fn, 'r') as fin:
            json_dicts.append(json.load(fin))
    sig = inspect.signature(func)
    kwargs['configs_filename'] = configs_filename
    kwargs = ChainMap(*([kwargs] + json_dicts))
    func_arg_keys = [k for k in sig.parameters]
    poped_keys = []
    func_arg_keys_tmp = list(func_arg_keys)
    for a, k in zip(args, func_arg_keys_tmp):
        poped_keys.append(func_arg_keys.pop(func_arg_keys.index(k)))
    kwargs_fine = {k: kwargs.get(k)
                   for k in func_arg_keys if kwargs.get(k) is not None}
    ba = sig.bind(*args, **kwargs_fine)
    ba.apply_defaults()
    return ba.args, ba.kwargs, poped_keys



""" Base class definition """
import tensorflow as tf
from dxpy.collections.dicts import DXDict

# On restrict mode, tensors as allocated to CPU memory if not specified.
RESTRICT_MODE = True


class Graph:
    """
    Base class of components.
    A graph is an abstraction of computational graph.
    It supports:
        1. Load configs externally. (By name(path))
        2. Exports some interface nodes.
        3. Supports some tasks, a dict of callables.
        4. A main tensor/graph which will be return by method as_tensor()
    """
    required_configs = []

    def __init__(name):
        from dxpy.collections import TreeDict
        self.name = name
        self.c = self._load_config()
        self._refine_config()
        self.nodes = dict()
        self.tasks = dict()
        self.main = None

    def _load_config(self):
        from ..config import config
        self.c = config[name]
        for k in self.required_configs:
            self.c[k] = config[name][k]

    def _refine_config(self):
        """
        Implement this method if additional configs is required.
        """
        pass

    def reigister_node(self, tensor_or_subgraph, name=None):
        if name is None:
            name = tensor_or_subgraph.name
        self.nodes[name] = tensor_or_subgraph

    def register_task(self, name, func):
        self.tasks[name] = func

    def __getitem__(self, name):
        if name in self.nodes:
            return self.nodes[name]
        elif name in self.tasks:
            return self.tasks[name]
        else:
            return None

    def __call__(self, feed_dict=None):
        return self.run('main', feed_dict)

    def run(self, task_name, feed_dict):
        return self.tasks[name](feed_dict)

    def as_tensor(self):
        return self.main


class Net(Graph):
    """ Base class of nets.
    Net add some special tasks based on graph:        
        1. train
        2. inference
        3. evaluate
        4. save/load
    """

    def __init__(self, name):
        super(__class__, self).__init__(name)

    def __getitem__(self, name):
        """
        Inputs:
            name: str, url like string.
        Returns:
            tensor or subnet.
        """
        from dxpy.filesystem.path import Path
        name = Path(name)
        if name.abs in self.nodes:
            return self.nodes[name.abs]
        else:
            names = name.parts
            return self.nodes[names[0]]['/'.join(names[1:])]
        return None

    def _run_sub_graph(self, sub_graph_name, feed_dict=None, sub_task_name=None):
        if sub_task_name is None:
            return self[sub_graph_name](feed_dict)
        else:
            return self[sub_graph_name].run(sub_task_name, feed_dict)

    def train(self, feed_dict=None, name=None):
        return self._run_sub_graph('train', feed_dict, name)

    def inference(self, feed_dict=None, name=None):
        return self._run_sub_graph('inference', feed_dict, name)

    def evaluate(self, feed_dict=None, name=None):
        return self._run_sub_graph('evaluate', feed_dict, name)

    def save(self, feed_dict=None):
        return self._run_sub_graph('saver', feed_dict, 'save')

    def load(self, feed_dict=None):
        return self._run_sub_graph('saver', feed_dict, 'load')



import click


@click.group()
def new():
    pass


@new.command()
@click.option('--name', '-n', type=str)
@click.option('--path', '-p', type=str, default='.')
def component(name, path):
    from . api import SnippetMaker
    SnippetMaker.component(name, path)


class CLI(click.MultiCommand):
    """
        Code maker
    """
    commands = {'new': None}

    def __init__(self):
        super(__class__, self).__init__(name='code', help=__class__.__doc__)

    def list_commands(self, ctx):
        return sorted(self.commands.keys())

    def get_command(self, ctx, name):
        if name in self.commands:
            if self.commands[name] is None:
                mapping = {
                    'new': new,
                }
                self.commands[name] = mapping.get(name)
        return self.commands.get(name)


main = CLI()



import tensorflow as tf


class TensorCreator:
    @classmethod
    def create(cls, name, shape, dtype, device=None):
        pass


class GPUTensorCreator:
    """
    A Helper class to create 
    """
    NB_UNNAMED = 0

    @classmethod
    def _get_name(cls, tensor, name=None):
        if name is not None:
            return name
        if tensor is None:
            result = "tensor_{0}".format(cls.NB_UNNAMED)
            cls.NB_UNNAMED += 1
            return result
        return tensor.name

    @classmethod
    def _get_shape(cls, tensor, shape=None):
        if shape is not None:
            return shape
        return tensor.shape.as_list()

    @classmethod
    def _get_dtype(cls, tensor, dtype=None):
        if dtype is not None:
            return dtype
        if tensor is not None:
            return tensor.dtype
        return tf.float32

    @classmethod
    def _get_shape_split(cls, shape, nb_gpus):
        if isinstance(shape, tf.TensorShape):
            shape_gpu = shape.as_list()
        else:
            shape_gpu = list(shape)
        shape_gpu[0] = shape[0] // nb_gpus
        return shape_gpu

    @classmethod
    def _check_campatblity(cls, tensor, shape=None):
        if tensor is None:
            return None
        tensor_shape = tensor.shape
        if len(tensor_shape) > 0:
            if tensor_shape is not None:
                tensor_shape = tensor_shape.as_list()
            if tensor_shape[0] is None:
                tensor_shape[0] = -1
            shape = list(shape)
            if shape[0] is None:
                shape[0] = -1
        if not tensor_shape == shape:
            if not (len(tensor_shape) == 0 and shape == [1]):
                raise ValueError(
                    "shape {0} differnt to tensor {1}.".format(shape, tensor))

    @classmethod
    def _create_main_tensor(cls, shape, dtype, device):
        if device is None:
            return tf.placeholder(dtype=dtype, shape=shape, name='main')
        else:
            with tf.device('/cpu:0'):
                return tf.placeholder(dtype=dtype, shape=shape, name='main')

    @classmethod
    def _slice_tensor_for_gpu(cls, tensor_input, nb_gpus):
        result = []
        shape_input = tensor_input.shape.as_list()
        shape_gpu = cls._get_shape_split(shape_input, nb_gpus)
        batch_size_gpu = shape_gpu[0]
        for id_slice in range(nb_gpus):
            with tf.device('/gpu:{}'.format(id_slice)):
                slice_start = [id_slice * batch_size_gpu] + \
                    [0] * (len(shape_gpu) - 1)
                result.append(tf.slice(tensor_input, slice_start,
                                       shape_gpu, name='part_{}'.format(id_slice)))
        if len(result) == 0:
            result = [tensor_input]
        return result

    @classmethod
    def _get_nb_gpu_and_device_main(cls, nb_gpu):
        if nb_gpu is None:
            nb_gpu = 0
            device_main = '/cpu:0'
        else:
            device_main = None
        return nb_gpu, device_main

    @classmethod
    def _get_tensor_main(cls, tensor, shape, dtype, device_main):
        if tensor is not None:
            return tensor
        else:
            return cls._create_main_tensor(shape, dtype, device_main)

    @classmethod
    def create(cls, tensor, name=None, shape=None, dtype=None, nb_gpu=None):
        name = cls._get_name(tensor, name)
        shape = cls._get_shape(tensor, shape)
        dtype = cls._get_dtype(tensor, dtype)
        cls._check_campatblity(tensor, shape)
        with tf.name_scope(name):
            nb_gpu, device_main = cls._get_nb_gpu_and_device_main(nb_gpu)
            tensor_main = cls._get_tensor_main(
                tensor, shape, dtype, device_main)
            tensors_gpu = cls._slice_tensor_for_gpu(tensor_main, nb_gpu)
        return tensor_main, tensors_gpu
